"""Dispatch plans for remote AI implementation via GitHub Actions."""

import logging
import tomllib
from dataclasses import dataclass
from datetime import UTC, datetime
from pathlib import Path

import click

from erk.cli.commands.pr.dispatch_helpers import ensure_trunk_synced
from erk.cli.commands.pr.metadata_helpers import write_dispatch_metadata
from erk.cli.commands.slot.common import is_placeholder_branch
from erk.cli.constants import (
    DISPATCH_WORKFLOW_METADATA_NAME,
    DISPATCH_WORKFLOW_NAME,
    ERK_PLAN_LABEL,
)
from erk.cli.core import discover_repo_context
from erk.cli.ensure import Ensure, UserFacingCliError
from erk.core.context import ErkContext
from erk.core.repo_discovery import RepoContext
from erk_shared.gateway.git.remote_ops.types import PushError
from erk_shared.gateway.github.metadata.core import (
    create_submission_queued_block,
    render_erk_issue_event,
)
from erk_shared.gateway.github.parsing import (
    construct_pr_url,
    construct_workflow_run_url,
    extract_owner_repo_from_github_url,
)
from erk_shared.gateway.github.types import PRNotFound
from erk_shared.impl_context import build_impl_context_files
from erk_shared.impl_folder import read_plan_ref, resolve_impl_dir
from erk_shared.output.output import user_output
from erk_shared.plan_store.types import PlanNotFound

logger = logging.getLogger(__name__)


def load_workflow_config(repo_root: Path, workflow_name: str) -> dict[str, str]:
    """Load workflow config from .erk/config.toml [workflows.<name>] section.

    Args:
        repo_root: Repository root path
        workflow_name: Workflow filename (with or without .yml/.yaml extension).
            Only the basename is used for config lookup.

    Returns:
        Dict of string key-value pairs for workflow inputs.
        Returns empty dict if config file or section doesn't exist.

    Example:
        For workflow_name="plan-implement.yml", reads from:
        .erk/config.toml -> [workflows.plan-implement] section
    """
    config_path = repo_root / ".erk" / "config.toml"

    if not config_path.exists():
        return {}

    with open(config_path, "rb") as f:
        data = tomllib.load(f)

    # Extract basename and strip .yml/.yaml extension
    basename = Path(workflow_name).name
    config_name = basename.removesuffix(".yml").removesuffix(".yaml")

    # Get [workflows.<name>] section
    workflows_section = data.get("workflows", {})
    workflow_config = workflows_section.get(config_name, {})

    # Convert all values to strings (workflow inputs are always strings)
    return {k: str(v) for k, v in workflow_config.items()}


@dataclass(frozen=True)
class DispatchResult:
    """Result of dispatching a single plan."""

    plan_number: int
    plan_title: str
    plan_url: str
    pr_number: int | None
    pr_url: str | None
    workflow_run_id: str
    workflow_url: str


def _build_workflow_run_url(plan_url: str, run_id: str) -> str:
    """Construct GitHub Actions workflow run URL from plan URL and run ID.

    Args:
        plan_url: GitHub plan URL (e.g., https://github.com/owner/repo/issues/123)
        run_id: Workflow run ID

    Returns:
        Workflow run URL (e.g., https://github.com/owner/repo/actions/runs/1234567890)
    """
    owner_repo = extract_owner_repo_from_github_url(plan_url)
    if owner_repo is not None:
        owner, repo = owner_repo
        return construct_workflow_run_url(owner, repo, run_id)
    return f"https://github.com/actions/runs/{run_id}"


def _build_pr_url(plan_url: str, pr_number: int) -> str:
    """Construct GitHub PR URL from plan URL and PR number.

    Args:
        plan_url: GitHub plan URL (e.g., https://github.com/owner/repo/issues/123)
        pr_number: PR number

    Returns:
        PR URL (e.g., https://github.com/owner/repo/pull/456)
    """
    owner_repo = extract_owner_repo_from_github_url(plan_url)
    if owner_repo is not None:
        owner, repo = owner_repo
        return construct_pr_url(owner, repo, pr_number)
    return f"https://github.com/pull/{pr_number}"


@dataclass(frozen=True)
class ValidatedPlannedPR:
    """Draft PR that passed all validation checks."""

    number: int
    title: str
    url: str
    branch_name: str


def _validate_planned_pr_for_dispatch(
    ctx: ErkContext,
    repo: RepoContext,
    plan_number: int,
) -> ValidatedPlannedPR:
    """Validate a planned PR plan for dispatch.

    Fetches the PR, validates it has the erk-plan label and is OPEN.

    Args:
        ctx: ErkContext with git operations
        repo: Repository context
        plan_number: PR number to validate

    Raises:
        SystemExit: If PR doesn't exist, missing label, or not OPEN.
    """
    pr_result = ctx.github.get_pr(repo.root, plan_number)
    if isinstance(pr_result, PRNotFound):
        user_output(click.style("Error: ", fg="red") + f"PR #{plan_number} not found")
        raise SystemExit(1)

    # Validate: must have erk-plan label
    if ERK_PLAN_LABEL not in pr_result.labels:
        user_output(
            click.style("Error: ", fg="red")
            + f"PR #{plan_number} does not have {ERK_PLAN_LABEL} label\n\n"
            "Cannot dispatch non-plan PRs for automated implementation."
        )
        raise SystemExit(1)

    # Validate: must be OPEN
    if pr_result.state != "OPEN":
        user_output(
            click.style("Error: ", fg="red") + f"PR #{plan_number} is {pr_result.state}\n\n"
            "Cannot dispatch closed PRs for automated implementation."
        )
        raise SystemExit(1)

    return ValidatedPlannedPR(
        number=plan_number,
        title=pr_result.title,
        url=pr_result.url,
        branch_name=pr_result.head_ref_name,
    )


def _dispatch_planned_pr_plan(
    ctx: ErkContext,
    *,
    repo: RepoContext,
    validated: ValidatedPlannedPR,
    submitted_by: str,
    base_branch: str,
    ref: str | None,
) -> DispatchResult:
    """Dispatch a validated planned-PR plan for implementation.

    For planned-PR plans, the branch and PR already exist. This function:
    - Syncs local branch ref to remote using git plumbing (no checkout)
    - Commits .erk/impl-context/ files directly to branch (no checkout)
    - Pushes and triggers the workflow with plan_backend="planned_pr"

    Args:
        ctx: ErkContext with git operations
        repo: Repository context
        validated: Validated planned PR information
        submitted_by: GitHub username of submitter
        base_branch: Base branch for PR

    Returns:
        DispatchResult with URLs and identifiers.
    """
    plan_number = validated.number
    branch_name = validated.branch_name

    # Fetch plan content via PlanBackend
    user_output("Fetching plan content...")
    result = ctx.plan_store.get_plan(repo.root, str(plan_number))
    if isinstance(result, PlanNotFound):
        user_output(click.style("Error: ", fg="red") + f"PR #{plan_number}: plan content not found")
        raise SystemExit(1)
    plan = result

    # Sync local branch ref to remote (no checkout required)
    user_output(f"Syncing branch: {click.style(branch_name, fg='cyan')}")
    ctx.git.remote.fetch_branch(repo.root, "origin", branch_name)
    ctx.git.branch.create_branch(repo.root, branch_name, f"origin/{branch_name}", force=True)

    # Commit impl-context files directly to branch (no checkout required)
    user_output("Committing plan to branch...")
    files = build_impl_context_files(
        plan_content=plan.body,
        plan_id=str(plan_number),
        url=validated.url,
        provider="github-draft-pr",
        objective_id=plan.objective_id,
        now_iso=ctx.time.now().isoformat(),
        node_ids=None,
    )
    ctx.git.commit.commit_files_to_branch(
        repo.root,
        branch=branch_name,
        files=files,
        message=f"Add plan for PR #{plan_number}",
    )
    push_result = ctx.git.remote.push_to_remote(
        repo.root, "origin", branch_name, set_upstream=False, force=False
    )
    if isinstance(push_result, PushError):
        raise UserFacingCliError(push_result.message)
    user_output(click.style("✓", fg="green") + " Branch pushed to remote")

    # Gather submission metadata
    queued_at = datetime.now(UTC).isoformat()

    # Load workflow-specific config
    workflow_config = load_workflow_config(repo.root, DISPATCH_WORKFLOW_NAME)

    # Build inputs dict with plan_backend="planned_pr"
    user_output("")
    user_output(f"Dispatching workflow: {click.style(DISPATCH_WORKFLOW_NAME, fg='cyan')}")

    inputs = {
        "plan_id": str(plan_number),
        "submitted_by": submitted_by,
        "plan_title": validated.title,
        "branch_name": branch_name,
        "pr_number": str(plan_number),
        "base_branch": base_branch,
        "plan_backend": "planned_pr",
        **workflow_config,
    }

    run_id = ctx.github.trigger_workflow(
        repo_root=repo.root,
        workflow=DISPATCH_WORKFLOW_NAME,
        inputs=inputs,
        ref=ref,
    )
    user_output(click.style("✓", fg="green") + " Workflow dispatched.")

    # Compute workflow URL
    workflow_url = _build_workflow_run_url(validated.url, run_id)

    # Write dispatch metadata FIRST (before any PR body modification)
    try:
        write_dispatch_metadata(
            plan_backend=ctx.plan_backend,
            github=ctx.github,
            repo_root=repo.root,
            plan_number=plan_number,
            run_id=run_id,
            dispatched_at=queued_at,
        )
        user_output(click.style("✓", fg="green") + " Dispatch metadata written")
    except Exception as e:
        user_output(
            click.style("Warning: ", fg="yellow") + f"Failed to update dispatch metadata: {e}"
        )

    # Update PR body with workflow run link (best-effort)
    # Guard: only append if body is non-empty to avoid overwriting metadata block
    try:
        pr_details = ctx.github.get_pr(repo.root, plan_number)
        if not isinstance(pr_details, PRNotFound) and pr_details.body:
            updated_body = pr_details.body + f"\n\n**Workflow run:** {workflow_url}"
            ctx.github.update_pr_body(repo.root, plan_number, updated_body)
    except Exception as e:
        logger.warning("Failed to update PR body with workflow run link: %s", e)

    # Post queued event comment via PlanBackend
    try:
        validation_results = {
            "pr_is_open": True,
            "has_erk_plan_label": True,
        }

        metadata_block = create_submission_queued_block(
            queued_at=queued_at,
            submitted_by=submitted_by,
            plan_number=plan_number,
            validation_results=validation_results,
            expected_workflow=DISPATCH_WORKFLOW_METADATA_NAME,
        )

        comment_body = render_erk_issue_event(
            title="Plan Queued for Implementation",
            metadata=metadata_block,
            description=(
                f"Plan submitted by **{submitted_by}** at {queued_at}.\n\n"
                f"The `{DISPATCH_WORKFLOW_METADATA_NAME}` workflow has been "
                f"dispatched via direct dispatch.\n\n"
                f"**Workflow run:** {workflow_url}"
            ),
        )

        user_output("Posting queued event comment...")
        ctx.plan_backend.add_comment(repo.root, str(plan_number), comment_body)
        user_output(click.style("✓", fg="green") + " Queued event comment posted")
    except Exception as e:
        user_output(
            click.style("Warning: ", fg="yellow")
            + f"Failed to post queued comment: {e}\n"
            + "Workflow is already running."
        )

    pr_url = _build_pr_url(validated.url, plan_number)

    return DispatchResult(
        plan_number=plan_number,
        plan_title=validated.title,
        plan_url=validated.url,
        pr_number=plan_number,
        pr_url=pr_url,
        workflow_run_id=run_id,
        workflow_url=workflow_url,
    )


def _detect_plan_number_from_context(
    repo: RepoContext,
    *,
    branch_name: str | None,
) -> int | None:
    """Detect plan PR number from local context when no argument given.

    Uses resolve_impl_dir() for unified discovery across .erk/impl-context/ directories.

    Args:
        repo: Repository context
        branch_name: Current git branch name, or None

    Returns:
        Detected PR number, or None if nothing found.
    """
    impl_dir = resolve_impl_dir(repo.root, branch_name=branch_name)
    if impl_dir is not None:
        plan_ref = read_plan_ref(impl_dir)
        if plan_ref is not None and plan_ref.plan_id.isdigit():
            return int(plan_ref.plan_id)

    return None


@click.command("dispatch")
@click.argument("plan_numbers", type=int, nargs=-1, required=False)
@click.option(
    "--base",
    type=str,
    default=None,
    help="Base branch for PR (defaults to current branch).",
)
@click.option(
    "--ref",
    "dispatch_ref",
    type=str,
    default=None,
    help="Branch to dispatch workflow from (overrides config dispatch_ref)",
)
@click.pass_obj
def pr_dispatch(
    ctx: ErkContext, plan_numbers: tuple[int, ...], base: str | None, dispatch_ref: str | None
) -> None:
    """Dispatch plans for remote AI implementation via GitHub Actions.

    Creates branch and draft PR locally (for correct commit attribution),
    then dispatches the plan-implement.yml GitHub Actions workflow.

    Arguments:
        PLAN_NUMBERS: One or more plan numbers to dispatch.
            If omitted, auto-detects from the resolved implementation directory or current branch.

    \b
    Example:
        erk pr dispatch 123
        erk pr dispatch 123 456 789
        erk pr dispatch 123 --base master
        erk pr dispatch                     # auto-detect from context

    Requires:
        - All issues must have erk-plan label
        - All issues must be OPEN
        - Working directory must be clean (no uncommitted changes)
    """
    ref = dispatch_ref if dispatch_ref is not None else ctx.local_config.dispatch_ref

    # Validate GitHub CLI prerequisites upfront (LBYL)
    user_output("Checking GitHub authentication...")
    Ensure.gh_authenticated(ctx)

    # Get repository context
    if isinstance(ctx.repo, RepoContext):
        repo = ctx.repo
    else:
        repo = discover_repo_context(ctx, ctx.cwd)

    # Ensure trunk is synced before any operations
    user_output("Syncing trunk with remote...")
    ensure_trunk_synced(ctx, repo)

    # Save current state (needed for both default base and restoration)
    original_branch = ctx.git.branch.get_current_branch(repo.root)
    if original_branch is None:
        user_output(
            click.style("Error: ", fg="red")
            + "Not on a branch (detached HEAD state). Cannot dispatch from here."
        )
        raise SystemExit(1)

    # If no arguments given, try to auto-detect from context
    if not plan_numbers:
        detected = _detect_plan_number_from_context(repo, branch_name=original_branch)
        if detected is None:
            user_output(
                click.style("Error: ", fg="red")
                + "No plan numbers provided and could not auto-detect from context.\n\n"
                "Provide plan numbers explicitly: erk pr dispatch <number>\n"
                "Or run from a plan branch with an associated PR."
            )
            raise SystemExit(1)
        user_output(f"Auto-detected PR #{detected} from context")
        plan_numbers = (detected,)

    # Validate base branch if provided, otherwise default to current branch (LBYL)
    if base is not None:
        if not ctx.git.branch.branch_exists_on_remote(repo.root, "origin", base):
            user_output(
                click.style("Error: ", fg="red") + f"Base branch '{base}' does not exist on remote"
            )
            raise SystemExit(1)
        target_branch = base
    else:
        # If on a placeholder branch (local-only), use trunk as base
        if is_placeholder_branch(original_branch):
            target_branch = ctx.git.branch.detect_trunk_branch(repo.root)
        elif not ctx.git.branch.branch_exists_on_remote(repo.root, "origin", original_branch):
            # Current branch not pushed to remote - fall back to trunk
            target_branch = ctx.git.branch.detect_trunk_branch(repo.root)
        else:
            target_branch = original_branch

    # Get GitHub username (authentication already validated)
    user_output("Resolving GitHub username...")
    _, username, _ = ctx.github.check_auth_status()
    submitted_by = username or "unknown"

    # Validate all planned-PR plans upfront
    user_output(f"Validating {len(plan_numbers)} planned-PR plan(s)...")
    user_output("")

    validated_planned_prs: list[ValidatedPlannedPR] = []
    for plan_number in plan_numbers:
        user_output(f"Validating PR #{plan_number}...")
        validated_pr = _validate_planned_pr_for_dispatch(ctx, repo, plan_number)
        validated_planned_prs.append(validated_pr)

    user_output("")
    user_output(
        click.style("✓", fg="green") + f" All {len(validated_planned_prs)} plan(s) validated"
    )
    user_output("")

    for v in validated_planned_prs:
        user_output(f"  #{v.number}: {click.style(v.title, fg='yellow')}")
    user_output("")

    # Dispatch all validated plans
    results: list[DispatchResult] = []
    for i, v in enumerate(validated_planned_prs):
        if len(validated_planned_prs) > 1:
            count = len(validated_planned_prs)
            user_output(f"--- Dispatching PR {i + 1}/{count}: #{v.number} ---")
        else:
            user_output(f"Dispatching PR #{v.number}...")
        user_output("")
        result = _dispatch_planned_pr_plan(
            ctx,
            repo=repo,
            validated=v,
            submitted_by=submitted_by,
            base_branch=target_branch,
            ref=ref,
        )
        results.append(result)
        user_output("")

    # Success output
    user_output("")
    user_output(click.style("✓", fg="green") + f" {len(results)} plan(s) dispatched successfully!")
    user_output("")
    user_output("Dispatched plans:")
    for r in results:
        user_output(f"  #{r.plan_number}: {r.plan_title}")
        user_output(f"    Plan: {r.plan_url}")
        if r.pr_url:
            user_output(f"    PR: {r.pr_url}")
        user_output(f"    Workflow: {r.workflow_url}")



"""Tests for erk exec objective-post-action-comment."""

import json
from datetime import UTC, datetime
from pathlib import Path

from click.testing import CliRunner

from erk.cli.commands.exec.scripts.objective_post_action_comment import (
    _format_action_comment,
    objective_post_action_comment,
)
from erk_shared.context.context import ErkContext
from erk_shared.gateway.github.fake import FakeGitHub
from erk_shared.gateway.github.issues.fake import FakeGitHubIssues
from erk_shared.gateway.github.issues.types import IssueInfo


def _make_issue(*, number: int) -> IssueInfo:
    return IssueInfo(
        number=number,
        title="My Objective",
        body="objective body",
        state="OPEN",
        url=f"https://github.com/owner/repo/issues/{number}",
        labels=["erk-objective"],
        assignees=[],
        created_at=datetime(2026, 1, 1, tzinfo=UTC),
        updated_at=datetime(2026, 1, 1, tzinfo=UTC),
        author="testuser",
    )


class TestFormatActionComment:
    def test_full_template(self) -> None:
        """Formats all sections correctly."""
        result = _format_action_comment(
            title="Implement auth system",
            date="2026-02-17",
            pr_number=6517,
            phase_step="1.1, 1.2",
            what_was_done=["Added user model", "Added JWT library"],
            lessons_learned=["Frozen dataclasses work well"],
            roadmap_updates=["Step 1.1: pending -> done", "Step 1.2: pending -> done"],
            body_reconciliation=[
                {"section": "Design Decisions", "change": "Updated auth approach"}
            ],
        )

        assert "## Action: Implement auth system" in result
        assert "**Date:** 2026-02-17" in result
        assert "**PR:** #6517" in result
        assert "**Phase/Step:** 1.1, 1.2" in result
        assert "### What Was Done" in result
        assert "- Added user model" in result
        assert "- Added JWT library" in result
        assert "### Lessons Learned" in result
        assert "- Frozen dataclasses work well" in result
        assert "### Roadmap Updates" in result
        assert "- Step 1.1: pending -> done" in result
        assert "### Body Reconciliation" in result
        assert "- **Design Decisions**: Updated auth approach" in result

    def test_no_body_reconciliation(self) -> None:
        """Omits Body Reconciliation section when empty."""
        result = _format_action_comment(
            title="Simple update",
            date="2026-02-17",
            pr_number=100,
            phase_step="1.1",
            what_was_done=["Did thing"],
            lessons_learned=["Learned thing"],
            roadmap_updates=["Step 1.1: pending -> done"],
            body_reconciliation=[],
        )

        assert "### Body Reconciliation" not in result

    def test_empty_lessons(self) -> None:
        """Handles empty lessons learned list."""
        result = _format_action_comment(
            title="Update",
            date="2026-02-17",
            pr_number=100,
            phase_step="1.1",
            what_was_done=["Did thing"],
            lessons_learned=[],
            roadmap_updates=["Step 1.1: pending -> done"],
            body_reconciliation=[],
        )

        assert "### Lessons Learned" in result
        # No bullet points after the header
        lessons_idx = result.index("### Lessons Learned")
        roadmap_idx = result.index("### Roadmap Updates")
        between = result[lessons_idx:roadmap_idx]
        assert "- " not in between


class TestObjectivePostActionComment:
    def test_happy_path(self, tmp_path: Path) -> None:
        """Posts formatted comment and returns success."""
        objective = _make_issue(number=6423)
        fake_issues = FakeGitHubIssues(issues={6423: objective})

        input_data = json.dumps(
            {
                "issue_number": 6423,
                "date": "2026-02-17",
                "pr_number": 6517,
                "phase_step": "1.1",
                "title": "Add user model",
                "what_was_done": ["Added user model"],
                "lessons_learned": ["Works well"],
                "roadmap_updates": ["Step 1.1: pending -> done"],
            }
        )

        runner = CliRunner()
        result = runner.invoke(
            objective_post_action_comment,
            [],
            input=input_data,
            obj=ErkContext.for_test(
                github=FakeGitHub(issues_gateway=fake_issues),
                repo_root=tmp_path,
                cwd=tmp_path,
            ),
        )

        assert result.exit_code == 0, result.output
        data = json.loads(result.output)
        assert data["success"] is True
        assert "comment_id" in data

        # Verify the comment was posted
        assert len(fake_issues.added_comments) == 1
        issue_number, body, _comment_id = fake_issues.added_comments[0]
        assert issue_number == 6423
        assert "## Action: Add user model" in body
        assert "**PR:** #6517" in body

    def test_missing_required_fields(self, tmp_path: Path) -> None:
        """Returns error when required fields are missing."""
        fake_issues = FakeGitHubIssues()

        input_data = json.dumps({"issue_number": 6423, "date": "2026-02-17"})

        runner = CliRunner()
        result = runner.invoke(
            objective_post_action_comment,
            [],
            input=input_data,
            obj=ErkContext.for_test(
                github=FakeGitHub(issues_gateway=fake_issues),
                repo_root=tmp_path,
                cwd=tmp_path,
            ),
        )

        assert result.exit_code == 1
        data = json.loads(result.output)
        assert data["success"] is False
        assert "Missing required fields" in data["error"]

    def test_empty_stdin(self, tmp_path: Path) -> None:
        """Returns error when stdin is empty."""
        fake_issues = FakeGitHubIssues()

        runner = CliRunner()
        result = runner.invoke(
            objective_post_action_comment,
            [],
            input="",
            obj=ErkContext.for_test(
                github=FakeGitHub(issues_gateway=fake_issues),
                repo_root=tmp_path,
                cwd=tmp_path,
            ),
        )

        assert result.exit_code == 1
        data = json.loads(result.output)
        assert data["success"] is False
        assert "No input" in data["error"]

    def test_with_body_reconciliation(self, tmp_path: Path) -> None:
        """Includes Body Reconciliation section when provided."""
        objective = _make_issue(number=6423)
        fake_issues = FakeGitHubIssues(issues={6423: objective})

        input_data = json.dumps(
            {
                "issue_number": 6423,
                "date": "2026-02-17",
                "pr_number": 6517,
                "phase_step": "1.1",
                "title": "Update approach",
                "what_was_done": ["Changed approach"],
                "body_reconciliation": [
                    {"section": "Design Decisions", "change": "Switched to WebSockets"}
                ],
            }
        )

        runner = CliRunner()
        result = runner.invoke(
            objective_post_action_comment,
            [],
            input=input_data,
            obj=ErkContext.for_test(
                github=FakeGitHub(issues_gateway=fake_issues),
                repo_root=tmp_path,
                cwd=tmp_path,
            ),
        )

        assert result.exit_code == 0, result.output
        _, body, _ = fake_issues.added_comments[0]
        assert "### Body Reconciliation" in body
        assert "**Design Decisions**: Switched to WebSockets" in body



"""Tests for erk pr rewrite command.

These tests verify the CLI layer behavior of the rewrite command.
The command squashes commits, generates an AI-powered commit message,
amends the commit, pushes, and updates the remote PR.
"""

from datetime import UTC, datetime

from click.testing import CliRunner

from erk.cli.commands.pr import pr_group
from erk_shared.gateway.branch_manager.types import SubmitBranchError
from erk_shared.gateway.git.fake import FakeGit
from erk_shared.gateway.github.fake import FakeGitHub
from erk_shared.gateway.github.issues.fake import FakeGitHubIssues
from erk_shared.gateway.github.issues.types import IssueInfo
from erk_shared.gateway.github.types import PRDetails
from erk_shared.gateway.graphite.fake import FakeGraphite
from erk_shared.gateway.graphite.types import BranchMetadata
from erk_shared.gateway.time.fake import FakeTime
from erk_shared.plan_store.planned_pr import PlannedPRBackend
from erk_shared.plan_store.planned_pr_lifecycle import build_plan_stage_body
from tests.fakes.prompt_executor import FakePromptExecutor
from tests.test_utils.context_builders import build_workspace_test_context
from tests.test_utils.env_helpers import ErkIsolatedFsEnv, erk_isolated_fs_env
from tests.test_utils.plan_helpers import format_plan_header_body_for_test


def _make_pr_details(
    *,
    number: int,
    branch: str,
    body: str = "",
) -> PRDetails:
    """Create a PRDetails for testing."""
    return PRDetails(
        number=number,
        url=f"https://github.com/owner/repo/pull/{number}",
        title="WIP",
        body=body,
        state="OPEN",
        is_draft=False,
        base_ref_name="main",
        head_ref_name=branch,
        is_cross_repository=False,
        mergeable="MERGEABLE",
        merge_state_status="CLEAN",
        owner="owner",
        repo="repo",
    )


def _make_rewrite_fakes(
    env: ErkIsolatedFsEnv,
    *,
    branch_name: str,
    pr_number: int = 42,
    pr_body: str = "",
    commits_ahead: int = 3,
    squash_raises: Exception | None = None,
    prompt_output: str = "Add awesome feature\n\nThis PR adds an awesome new feature.",
    prompt_error: str | None = None,
    submit_branch_error: SubmitBranchError | None = None,
) -> tuple[FakeGit, FakeGraphite, FakeGitHub, FakePromptExecutor]:
    """Create standard fakes for rewrite command tests."""
    pr_details = _make_pr_details(number=pr_number, branch=branch_name, body=pr_body)

    git = FakeGit(
        git_common_dirs={env.cwd: env.git_dir},
        repository_roots={env.cwd: env.git_dir},
        local_branches={env.cwd: ["main", branch_name]},
        default_branches={env.cwd: "main"},
        trunk_branches={env.git_dir: "main"},
        current_branches={env.cwd: branch_name},
        commits_ahead={(env.cwd, "main"): commits_ahead},
        diff_to_branch={(env.cwd, "main"): "diff --git a/file.py b/file.py\n+new content"},
    )

    graphite = FakeGraphite(
        authenticated=True,
        squash_branch_raises=squash_raises,
        branches={
            branch_name: BranchMetadata(
                name=branch_name,
                parent="main",
                children=[],
                is_trunk=False,
                commit_sha=None,
            ),
            "main": BranchMetadata(
                name="main",
                parent=None,
                children=[branch_name],
                is_trunk=True,
                commit_sha=None,
            ),
        },
    )

    github = FakeGitHub(
        authenticated=True,
        prs_by_branch={branch_name: pr_details},
    )

    executor_kwargs: dict = {"available": True}
    if prompt_error is not None:
        executor_kwargs["simulated_prompt_error"] = prompt_error
    else:
        executor_kwargs["simulated_prompt_output"] = prompt_output
    executor = FakePromptExecutor(**executor_kwargs)

    return git, graphite, github, executor


def test_pr_rewrite_happy_path() -> None:
    """Test successful rewrite: squash, generate, amend, push, update PR."""
    runner = CliRunner()
    with erk_isolated_fs_env(runner, env_overrides=None) as env:
        git, graphite, github, executor = _make_rewrite_fakes(env, branch_name="feature")

        ctx = build_workspace_test_context(
            env,
            git=git,
            github=github,
            graphite=graphite,
            prompt_executor=executor,
        )

        result = runner.invoke(pr_group, ["rewrite"], obj=ctx)

        assert result.exit_code == 0, result.output
        assert "PR title and description updated" in result.output
        assert "Add awesome feature" in result.output

        # Verify commit was amended
        assert len(git.commits) == 1
        commit_message = git.commits[0].message
        assert "Add awesome feature" in commit_message
        assert "awesome new feature" in commit_message

        # Verify branch was pushed (Graphite submit_stack was called)
        assert "Branch pushed" in result.output

        # Verify PR was updated
        assert len(github.updated_pr_titles) == 1
        assert github.updated_pr_titles[0] == (42, "Add awesome feature")


def test_pr_rewrite_already_single_commit() -> None:
    """Test rewrite with already-single-commit (squash is idempotent)."""
    runner = CliRunner()
    with erk_isolated_fs_env(runner, env_overrides=None) as env:
        git, graphite, github, executor = _make_rewrite_fakes(
            env, branch_name="feature", commits_ahead=1
        )

        ctx = build_workspace_test_context(
            env,
            git=git,
            github=github,
            graphite=graphite,
            prompt_executor=executor,
        )

        result = runner.invoke(pr_group, ["rewrite"], obj=ctx)

        assert result.exit_code == 0, result.output
        assert "PR title and description updated" in result.output


def test_pr_rewrite_fails_when_no_pr() -> None:
    """Test that rewrite fails when no PR exists for the branch."""
    runner = CliRunner()
    with erk_isolated_fs_env(runner, env_overrides=None) as env:
        git = FakeGit(
            git_common_dirs={env.cwd: env.git_dir},
            repository_roots={env.cwd: env.git_dir},
            local_branches={env.cwd: ["main", "feature"]},
            default_branches={env.cwd: "main"},
            trunk_branches={env.git_dir: "main"},
            current_branches={env.cwd: "feature"},
            commits_ahead={(env.cwd, "main"): 1},
        )

        graphite = FakeGraphite(
            authenticated=True,
            branches={
                "feature": BranchMetadata(
                    name="feature",
                    parent="main",
                    children=[],
                    is_trunk=False,
                    commit_sha=None,
                ),
                "main": BranchMetadata(
                    name="main",
                    parent=None,
                    children=["feature"],
                    is_trunk=True,
                    commit_sha=None,
                ),
            },
        )
        # No PR configured for the branch
        github = FakeGitHub(authenticated=True)
        executor = FakePromptExecutor(available=True)

        ctx = build_workspace_test_context(
            env,
            git=git,
            github=github,
            graphite=graphite,
            prompt_executor=executor,
        )

        result = runner.invoke(pr_group, ["rewrite"], obj=ctx)

        assert result.exit_code != 0
        assert "No PR found" in result.output
        assert "erk pr submit" in result.output


def test_pr_rewrite_fails_when_squash_conflicts() -> None:
    """Test that rewrite fails when squash encounters conflicts."""
    runner = CliRunner()
    with erk_isolated_fs_env(runner, env_overrides=None) as env:
        git, graphite, github, executor = _make_rewrite_fakes(
            env,
            branch_name="feature",
            squash_raises=RuntimeError("Squash conflict: resolve manually"),
        )

        ctx = build_workspace_test_context(
            env,
            git=git,
            github=github,
            graphite=graphite,
            prompt_executor=executor,
        )

        result = runner.invoke(pr_group, ["rewrite"], obj=ctx)

        assert result.exit_code != 0
        assert "Squash failed" in result.output


def test_pr_rewrite_fails_when_ai_generation_fails() -> None:
    """Test that rewrite fails when commit message generation fails."""
    runner = CliRunner()
    with erk_isolated_fs_env(runner, env_overrides=None) as env:
        git, graphite, github, executor = _make_rewrite_fakes(
            env,
            branch_name="feature",
            prompt_error="Claude CLI execution failed",
        )

        ctx = build_workspace_test_context(
            env,
            git=git,
            github=github,
            graphite=graphite,
            prompt_executor=executor,
        )

        result = runner.invoke(pr_group, ["rewrite"], obj=ctx)

        assert result.exit_code != 0
        assert "Failed to generate message" in result.output


def test_pr_rewrite_fails_when_detached_head() -> None:
    """Test that rewrite fails when in detached HEAD state."""
    runner = CliRunner()
    with erk_isolated_fs_env(runner, env_overrides=None) as env:
        git = FakeGit(
            git_common_dirs={env.cwd: env.git_dir},
            local_branches={env.cwd: ["main"]},
            default_branches={env.cwd: "main"},
            # No current_branches means detached HEAD (get_current_branch returns None)
        )
        executor = FakePromptExecutor(available=True)

        ctx = build_workspace_test_context(env, git=git, prompt_executor=executor)

        result = runner.invoke(pr_group, ["rewrite"], obj=ctx)

        assert result.exit_code != 0
        assert "detached HEAD" in result.output


def test_pr_rewrite_fails_when_claude_not_available() -> None:
    """Test that rewrite fails when Claude CLI is not available."""
    runner = CliRunner()
    with erk_isolated_fs_env(runner, env_overrides=None) as env:
        git = FakeGit(
            git_common_dirs={env.cwd: env.git_dir},
            local_branches={env.cwd: ["main"]},
            default_branches={env.cwd: "main"},
        )
        executor = FakePromptExecutor(available=False)

        ctx = build_workspace_test_context(env, git=git, prompt_executor=executor)

        result = runner.invoke(pr_group, ["rewrite"], obj=ctx)

        assert result.exit_code != 0
        assert "Claude CLI not found" in result.output


def test_pr_rewrite_planned_pr_backend_preserves_metadata() -> None:
    """Test that rewrite preserves metadata prefix and omits Closes # for draft PR backend."""
    runner = CliRunner()
    with erk_isolated_fs_env(runner, env_overrides=None) as env:
        metadata_body = format_plan_header_body_for_test()
        plan_content = "# My Plan\n\nImplement the thing."
        pr_body = build_plan_stage_body(metadata_body, plan_content, summary=None)

        git, graphite, github, executor = _make_rewrite_fakes(
            env,
            branch_name="feature",
            pr_body=pr_body,
        )

        ctx = build_workspace_test_context(
            env,
            git=git,
            github=github,
            graphite=graphite,
            prompt_executor=executor,
            plan_store=PlannedPRBackend(github, github.issues, time=FakeTime()),
        )

        result = runner.invoke(pr_group, ["rewrite"], obj=ctx)

        assert result.exit_code == 0, result.output

        # Verify the PR body preserves metadata prefix
        assert len(github.updated_pr_bodies) == 1
        updated_body = github.updated_pr_bodies[0][1]
        assert "plan-header" in updated_body
        # No self-closing reference for draft PR backend
        assert "Closes #" not in updated_body


def test_pr_rewrite_retracks_graphite_branch_after_amend() -> None:
    """Test that rewrite retracks Graphite branch immediately after amending commit.

    The retrack must happen right after amend (before push or PR update) so that
    Graphite's cache stays consistent even if subsequent operations fail.
    """
    runner = CliRunner()
    with erk_isolated_fs_env(runner, env_overrides=None) as env:
        git, graphite, github, executor = _make_rewrite_fakes(env, branch_name="feature")

        ctx = build_workspace_test_context(
            env,
            git=git,
            github=github,
            graphite=graphite,
            prompt_executor=executor,
            use_graphite=True,
        )

        result = runner.invoke(pr_group, ["rewrite"], obj=ctx)

        assert result.exit_code == 0, result.output
        # Verify retrack_branch was called with repo root and branch name
        assert len(graphite.retrack_branch_calls) == 1
        retrack_cwd, retrack_branch = graphite.retrack_branch_calls[0]
        assert retrack_cwd == env.git_dir
        assert retrack_branch == "feature"


def test_pr_rewrite_skips_lifecycle_when_plan_not_resolved() -> None:
    """Rewrite skips lifecycle update when plan cannot be resolved from branch name."""
    runner = CliRunner()
    with erk_isolated_fs_env(runner, env_overrides=None) as env:
        # plnd/ branch — resolve_plan_id_for_branch returns None
        branch_name = "plnd/add-feature"

        # Create plan issue with lifecycle_stage "planned"
        plan_body = format_plan_header_body_for_test(lifecycle_stage="planned")
        plan_issue = IssueInfo(
            number=100,
            title="Plan #100",
            body=plan_body,
            state="OPEN",
            url="https://github.com/owner/repo/issues/100",
            labels=["erk-pr", "erk-plan"],
            assignees=[],
            created_at=datetime(2024, 1, 15, 10, 30, tzinfo=UTC),
            updated_at=datetime(2024, 1, 15, 10, 30, tzinfo=UTC),
            author="test-user",
        )
        fake_issues = FakeGitHubIssues(issues={100: plan_issue})

        git, graphite, github, executor = _make_rewrite_fakes(
            env,
            branch_name=branch_name,
        )

        ctx = build_workspace_test_context(
            env,
            git=git,
            github=github,
            graphite=graphite,
            prompt_executor=executor,
            issues=fake_issues,
        )

        result = runner.invoke(pr_group, ["rewrite"], obj=ctx)

        assert result.exit_code == 0, result.output

        # Plan cannot be resolved from branch name — no lifecycle update
        assert len(fake_issues.updated_bodies) == 0



"""Filter action handlers for TUI application."""

from __future__ import annotations

from typing import TYPE_CHECKING

from erk.tui.filtering.types import FilterState
from erk.tui.sorting.types import BranchActivity, SortKey

if TYPE_CHECKING:
    from erk.tui.app import ErkDashApp


class FilterActionsMixin:
    """Mixin providing filter and sort action handlers.

    Includes action_start_filter, action_toggle_stack_filter,
    action_toggle_objective_filter, action_toggle_sort,
    and related helpers.
    """

    def action_start_filter(self: ErkDashApp) -> None:
        """Activate filter mode and focus the input."""
        if self._filter_input is None:
            return
        self._filter_state = self._filter_state.activate()
        self._filter_input.disabled = False
        self._filter_input.add_class("visible")
        self._filter_input.focus()

    def action_toggle_stack_filter(self: ErkDashApp) -> None:
        """Toggle stack filter for the focused row's Graphite stack.

        If stack filter is active, clears it (toggle off).
        If no row selected or row has no branch, shows status message.
        If branch is not in a Graphite stack, shows status message.
        Otherwise, filters table to only show rows in the same stack.
        """
        # Toggle off if already active
        if self._stack_filter_branches is not None:
            self._clear_stack_filter()
            return

        row = self._get_selected_row()
        if row is None or row.pr_head_branch is None:
            if self._status_bar is not None:
                self._status_bar.set_message("No branch for stack filter")
            return

        stack = self._provider.get_branch_stack(row.pr_head_branch)
        if stack is None:
            if self._status_bar is not None:
                self._status_bar.set_message("Not in a Graphite stack")
            return

        self._stack_filter_branches = frozenset(stack)
        self._stack_filter_label = row.pr_head_branch
        self._apply_filter()

        if self._status_bar is not None:
            self._status_bar.set_message(f"Stack: {row.pr_head_branch} ({len(stack)})")

    def _clear_stack_filter(self: ErkDashApp) -> None:
        """Clear the stack filter and restore all rows."""
        self._stack_filter_branches = None
        self._stack_filter_label = None
        self._apply_filter()
        if self._status_bar is not None:
            self._status_bar.set_message(None)

    def action_toggle_objective_filter(self: ErkDashApp) -> None:
        """Toggle objective filter for the focused row's objective.

        If objective filter is active, clears it (toggle off).
        If no row selected or row has no objective_issue, shows status message.
        Otherwise, filters table to only show rows sharing the same objective.
        """
        # Toggle off if already active
        if self._objective_filter_issue is not None:
            self._clear_objective_filter()
            return

        row = self._get_selected_row()
        if row is None:
            return

        if row.objective_issue is None:
            if self._status_bar is not None:
                self._status_bar.set_message("Plan not linked to an objective")
            return

        self._objective_filter_issue = row.objective_issue
        self._objective_filter_label = str(row.objective_issue)
        self._apply_filter()

        count = len(self._rows)
        if self._status_bar is not None:
            self._status_bar.set_message(f"Objective: #{row.objective_issue} ({count})")

    def _clear_objective_filter(self: ErkDashApp) -> None:
        """Clear the objective filter and restore all rows."""
        self._objective_filter_issue = None
        self._objective_filter_label = None
        self._apply_filter()
        if self._status_bar is not None:
            self._status_bar.set_message(None)

    def action_toggle_sort(self: ErkDashApp) -> None:
        """Toggle between sort modes."""
        self._sort_state = self._sort_state.toggle()

        # If switching to activity sort, load activity data in background
        if self._sort_state.key == SortKey.BRANCH_ACTIVITY and not self._activity_by_plan:
            self._load_activity_and_resort()
        else:
            # Re-sort with current data
            self._rows = self._apply_filter_and_sort(self._all_rows)
            if self._table is not None:
                self._table.populate(self._rows)

        # Update status bar
        if self._status_bar is not None:
            self._status_bar.set_sort_mode(self._sort_state.display_label)

    def _on_activity_loaded(self: ErkDashApp, activity: dict[int, BranchActivity]) -> None:
        """Handle activity data loaded - resort the table."""
        self._activity_by_plan = activity
        self._activity_loading = False

        # Re-sort with new activity data
        self._rows = self._apply_filter_and_sort(self._all_rows)
        if self._table is not None:
            self._table.populate(self._rows)

    def _apply_filter(self: ErkDashApp) -> None:
        """Apply current filter query to the table."""
        self._rows = self._apply_filter_and_sort(self._all_rows)

        if self._table is not None:
            self._table.populate(self._rows)

        if self._status_bar is not None:
            self._status_bar.set_plan_count(
                len(self._rows), noun=self._display_name_for_view(self._view_mode).lower()
            )

    def _exit_filter_mode(self: ErkDashApp) -> None:
        """Exit filter mode, restore all rows, and focus table."""
        if self._filter_input is not None:
            self._filter_input.value = ""
            self._filter_input.remove_class("visible")
            self._filter_input.disabled = True

        self._filter_state = FilterState.initial()
        self._rows = self._apply_filter_and_sort(self._all_rows)

        if self._table is not None:
            self._table.populate(self._rows)
            self._table.focus()

        if self._status_bar is not None:
            self._status_bar.set_plan_count(
                len(self._rows), noun=self._display_name_for_view(self._view_mode).lower()
            )



"""Unit tests for detect_installed_backends."""

from erk.cli.commands.init.main import detect_installed_backends
from tests.fakes.shell import FakeShell


def test_detect_both_backends() -> None:
    """Both claude and codex installed returns both."""
    shell = FakeShell(installed_tools={"claude": "/usr/bin/claude", "codex": "/usr/bin/codex"})
    result = detect_installed_backends(shell)
    assert result == ["claude", "codex"]


def test_detect_only_claude() -> None:
    """Only claude installed returns just claude."""
    shell = FakeShell(installed_tools={"claude": "/usr/bin/claude"})
    result = detect_installed_backends(shell)
    assert result == ["claude"]


def test_detect_only_codex() -> None:
    """Only codex installed returns just codex."""
    shell = FakeShell(installed_tools={"codex": "/usr/bin/codex"})
    result = detect_installed_backends(shell)
    assert result == ["codex"]


def test_detect_no_backends() -> None:
    """No backends installed returns empty list."""
    shell = FakeShell(installed_tools={})
    result = detect_installed_backends(shell)
    assert result == []



"""Tests for plan close command."""

from datetime import UTC, datetime

from click.testing import CliRunner

from erk.cli.cli import cli
from erk_shared.gateway.github.fake import FakeGitHub
from erk_shared.gateway.github.issues.fake import FakeGitHubIssues
from erk_shared.gateway.github.issues.types import IssueInfo, PRReference
from erk_shared.gateway.time.fake import FakeTime
from erk_shared.plan_store.planned_pr import PlannedPRBackend
from erk_shared.plan_store.types import Plan, PlanState
from tests.fakes.prompt_executor import FakePromptExecutor
from tests.test_utils.context_builders import build_workspace_test_context
from tests.test_utils.env_helpers import erk_inmem_env
from tests.test_utils.plan_helpers import (
    create_plan_store_with_plans,
    format_plan_header_body_for_test,
    issue_info_to_pr_details,
)


def test_close_plan_with_plan_number() -> None:
    """Test closing a plan with plan number."""
    # Arrange
    plan_issue = Plan(
        plan_identifier="42",
        title="Test Issue",
        body="This is a test issue",
        state=PlanState.OPEN,
        url="https://github.com/owner/repo/issues/42",
        labels=["erk-pr", "erk-plan"],
        assignees=[],
        created_at=datetime(2024, 1, 1, tzinfo=UTC),
        updated_at=datetime(2024, 1, 2, tzinfo=UTC),
        metadata={},
        objective_id=None,
    )

    runner = CliRunner()
    with erk_inmem_env(runner) as env:
        store, fake_github = create_plan_store_with_plans({"42": plan_issue})
        ctx = build_workspace_test_context(env, plan_store=store, issues=fake_github.issues)

        # Act
        result = runner.invoke(cli, ["pr", "close", "42"], obj=ctx)

        # Assert
        assert result.exit_code == 0
        assert "Closed plan #42" in result.output
        assert 42 in fake_github.closed_prs
        # Verify PlannedPRBackend added a comment before closing
        assert any(num == 42 and "completed" in body for num, body in fake_github.pr_comments)


def test_close_plan_not_found() -> None:
    """Test closing a plan that doesn't exist."""
    # Arrange
    runner = CliRunner()
    with erk_inmem_env(runner) as env:
        store, _ = create_plan_store_with_plans({})
        ctx = build_workspace_test_context(env, plan_store=store)

        # Act
        result = runner.invoke(cli, ["pr", "close", "999"], obj=ctx)

        # Assert
        assert result.exit_code == 1
        assert "Error" in result.output
        assert "Plan #999 not found" in result.output


def _make_issue_info(plan: Plan) -> IssueInfo:
    """Helper to convert Plan to IssueInfo for tests needing custom FakeGitHubIssues config."""
    state = "OPEN" if plan.state == PlanState.OPEN else "CLOSED"
    return IssueInfo(
        number=int(plan.plan_identifier),
        title=plan.title,
        body=plan.body,
        state=state,
        url=plan.url,
        labels=plan.labels,
        assignees=plan.assignees,
        created_at=plan.created_at.astimezone(UTC),
        updated_at=plan.updated_at.astimezone(UTC),
        author="test-author",
    )


def test_close_plan_closes_linked_open_prs() -> None:
    """Test closing a plan closes all OPEN PRs linked to the issue."""
    # Arrange
    plan_issue = Plan(
        plan_identifier="42",
        title="Test Issue",
        body="This is a test issue",
        state=PlanState.OPEN,
        url="https://github.com/owner/repo/issues/42",
        labels=["erk-pr", "erk-plan"],
        assignees=[],
        created_at=datetime(2024, 1, 1, tzinfo=UTC),
        updated_at=datetime(2024, 1, 2, tzinfo=UTC),
        metadata={},
        objective_id=None,
    )

    # Create linked PRs (one draft, one non-draft, both OPEN)
    open_draft_pr = PRReference(number=100, state="OPEN", is_draft=True)
    open_non_draft_pr = PRReference(number=101, state="OPEN", is_draft=False)

    runner = CliRunner()
    with erk_inmem_env(runner) as env:
        issue = _make_issue_info(plan_issue)
        # Create FakeGitHubIssues with both the plan issue and PR references
        fake_issues = FakeGitHubIssues(
            issues={42: issue},
            pr_references={42: [open_draft_pr, open_non_draft_pr]},
        )
        fake_github = FakeGitHub(
            pr_details={42: issue_info_to_pr_details(issue)},
            issues_gateway=fake_issues,
        )
        store = PlannedPRBackend(fake_github, fake_issues, time=FakeTime())
        ctx = build_workspace_test_context(
            env, plan_store=store, github=fake_github, issues=fake_issues
        )

        # Act
        result = runner.invoke(cli, ["pr", "close", "42"], obj=ctx)

        # Assert
        assert result.exit_code == 0
        assert "Closed plan #42" in result.output
        assert "Closed 2 linked PR(s): #100, #101" in result.output
        # Verify both linked PRs were closed via FakeGitHub
        assert 100 in fake_github.closed_prs
        assert 101 in fake_github.closed_prs


def test_close_plan_skips_closed_and_merged_prs() -> None:
    """Test closing a plan skips CLOSED and MERGED PRs, only closes OPEN."""
    # Arrange
    plan_issue = Plan(
        plan_identifier="42",
        title="Test Issue",
        body="This is a test issue",
        state=PlanState.OPEN,
        url="https://github.com/owner/repo/issues/42",
        labels=["erk-pr", "erk-plan"],
        assignees=[],
        created_at=datetime(2024, 1, 1, tzinfo=UTC),
        updated_at=datetime(2024, 1, 2, tzinfo=UTC),
        metadata={},
        objective_id=None,
    )

    # Create PRs in various states
    open_pr = PRReference(number=100, state="OPEN", is_draft=False)
    closed_pr = PRReference(number=101, state="CLOSED", is_draft=False)
    merged_pr = PRReference(number=102, state="MERGED", is_draft=False)

    runner = CliRunner()
    with erk_inmem_env(runner) as env:
        issue = _make_issue_info(plan_issue)
        fake_issues = FakeGitHubIssues(
            issues={42: issue},
            pr_references={42: [open_pr, closed_pr, merged_pr]},
        )
        fake_github = FakeGitHub(
            pr_details={42: issue_info_to_pr_details(issue)},
            issues_gateway=fake_issues,
        )
        store = PlannedPRBackend(fake_github, fake_issues, time=FakeTime())
        ctx = build_workspace_test_context(
            env, plan_store=store, github=fake_github, issues=fake_issues
        )

        # Act
        result = runner.invoke(cli, ["pr", "close", "42"], obj=ctx)

        # Assert
        assert result.exit_code == 0
        assert "Closed plan #42" in result.output
        assert "Closed 1 linked PR(s): #100" in result.output
        # Only the OPEN linked PR should be in the linked-PR closed output
        assert 100 in fake_github.closed_prs
        assert 101 not in fake_github.closed_prs
        assert 102 not in fake_github.closed_prs


def test_close_plan_no_linked_prs() -> None:
    """Test closing a plan with no linked PRs works without error."""
    # Arrange
    plan_issue = Plan(
        plan_identifier="42",
        title="Test Issue",
        body="This is a test issue",
        state=PlanState.OPEN,
        url="https://github.com/owner/repo/issues/42",
        labels=["erk-pr", "erk-plan"],
        assignees=[],
        created_at=datetime(2024, 1, 1, tzinfo=UTC),
        updated_at=datetime(2024, 1, 2, tzinfo=UTC),
        metadata={},
        objective_id=None,
    )

    runner = CliRunner()
    with erk_inmem_env(runner) as env:
        issue = _make_issue_info(plan_issue)
        fake_issues = FakeGitHubIssues(
            issues={42: issue},
            pr_references={},  # No linked PRs
        )
        fake_github = FakeGitHub(
            pr_details={42: issue_info_to_pr_details(issue)},
            issues_gateway=fake_issues,
        )
        store = PlannedPRBackend(fake_github, fake_issues, time=FakeTime())
        ctx = build_workspace_test_context(
            env, plan_store=store, github=fake_github, issues=fake_issues
        )

        # Act
        result = runner.invoke(cli, ["pr", "close", "42"], obj=ctx)

        # Assert
        assert result.exit_code == 0
        assert "Closed plan #42" in result.output
        # No linked PR closing message should appear
        assert "linked PR(s)" not in result.output


def test_close_plan_invalid_identifier() -> None:
    """Test closing a plan with invalid identifier fails with helpful error."""
    # Arrange
    runner = CliRunner()
    with erk_inmem_env(runner) as env:
        store, _ = create_plan_store_with_plans({})
        ctx = build_workspace_test_context(env, plan_store=store)

        # Act
        result = runner.invoke(cli, ["pr", "close", "not-a-number"], obj=ctx)

        # Assert
        assert result.exit_code != 0
        assert "Invalid plan number or URL" in result.output
        assert "not-a-number" in result.output


def test_close_plan_invalid_url_format() -> None:
    """Test closing a plan with invalid URL format gives specific error."""
    # Arrange
    runner = CliRunner()
    with erk_inmem_env(runner) as env:
        store, _ = create_plan_store_with_plans({})
        ctx = build_workspace_test_context(env, plan_store=store)

        # Act - GitHub URL but pointing to pulls instead of issues
        result = runner.invoke(
            cli, ["pr", "close", "https://github.com/owner/repo/pulls/42"], obj=ctx
        )

        # Assert
        assert result.exit_code != 0
        assert "Invalid plan number or URL" in result.output
        assert "https://github.com/owner/repo/issues/456" in result.output


def test_close_plan_reports_closed_prs() -> None:
    """Test closing a plan reports the closed PRs in output."""
    # Arrange
    plan_issue = Plan(
        plan_identifier="42",
        title="Test Issue",
        body="This is a test issue",
        state=PlanState.OPEN,
        url="https://github.com/owner/repo/issues/42",
        labels=["erk-pr", "erk-plan"],
        assignees=[],
        created_at=datetime(2024, 1, 1, tzinfo=UTC),
        updated_at=datetime(2024, 1, 2, tzinfo=UTC),
        metadata={},
        objective_id=None,
    )

    # Create multiple linked OPEN PRs
    pr1 = PRReference(number=100, state="OPEN", is_draft=False)
    pr2 = PRReference(number=200, state="OPEN", is_draft=False)
    pr3 = PRReference(number=300, state="OPEN", is_draft=False)

    runner = CliRunner()
    with erk_inmem_env(runner) as env:
        issue = _make_issue_info(plan_issue)
        fake_issues = FakeGitHubIssues(
            issues={42: issue},
            pr_references={42: [pr1, pr2, pr3]},
        )
        fake_github = FakeGitHub(
            pr_details={42: issue_info_to_pr_details(issue)},
            issues_gateway=fake_issues,
        )
        store = PlannedPRBackend(fake_github, fake_issues, time=FakeTime())
        ctx = build_workspace_test_context(
            env, plan_store=store, github=fake_github, issues=fake_issues
        )

        # Act
        result = runner.invoke(cli, ["pr", "close", "42"], obj=ctx)

        # Assert
        assert result.exit_code == 0
        assert "Closed 3 linked PR(s): #100, #200, #300" in result.output


def test_close_plan_with_objective_invokes_update() -> None:
    """Test closing a plan linked to an objective invokes the objective update."""
    # Arrange - body must include plan-header with objective_issue so
    # PlannedPRBackend._convert_to_plan() extracts objective_id correctly
    body_with_header = format_plan_header_body_for_test(objective_issue=99) + "\nPlan content"
    plan_issue = Plan(
        plan_identifier="42",
        title="Test Issue",
        body=body_with_header,
        state=PlanState.OPEN,
        url="https://github.com/owner/repo/issues/42",
        labels=["erk-pr", "erk-plan"],
        assignees=[],
        created_at=datetime(2024, 1, 1, tzinfo=UTC),
        updated_at=datetime(2024, 1, 2, tzinfo=UTC),
        metadata={"issue_body": body_with_header},
        objective_id=99,
    )

    runner = CliRunner()
    with erk_inmem_env(runner) as env:
        executor = FakePromptExecutor()
        store, fake_github = create_plan_store_with_plans({"42": plan_issue})
        ctx = build_workspace_test_context(
            env, plan_store=store, issues=fake_github.issues, prompt_executor=executor
        )

        # Act
        result = runner.invoke(cli, ["pr", "close", "42"], obj=ctx)

        # Assert
        assert result.exit_code == 0
        assert "Closed plan #42" in result.output
        # Verify the objective update command was invoked
        assert len(executor.executed_commands) == 1
        executed_cmd = executor.executed_commands[0][0]
        assert "/erk:objective-update-with-closed-plan" in executed_cmd
        assert "--plan 42" in executed_cmd
        assert "--objective 99" in executed_cmd


def test_close_plan_without_objective_skips_update() -> None:
    """Test closing a plan without an objective does not invoke objective update."""
    # Arrange
    plan_issue = Plan(
        plan_identifier="42",
        title="Test Issue",
        body="This is a test issue",
        state=PlanState.OPEN,
        url="https://github.com/owner/repo/issues/42",
        labels=["erk-pr", "erk-plan"],
        assignees=[],
        created_at=datetime(2024, 1, 1, tzinfo=UTC),
        updated_at=datetime(2024, 1, 2, tzinfo=UTC),
        metadata={},
        objective_id=None,
    )

    runner = CliRunner()
    with erk_inmem_env(runner) as env:
        executor = FakePromptExecutor()
        store, fake_github = create_plan_store_with_plans({"42": plan_issue})
        ctx = build_workspace_test_context(
            env, plan_store=store, issues=fake_github.issues, prompt_executor=executor
        )

        # Act
        result = runner.invoke(cli, ["pr", "close", "42"], obj=ctx)

        # Assert
        assert result.exit_code == 0
        assert "Closed plan #42" in result.output
        # No objective update should have been invoked
        assert len(executor.executed_commands) == 0


def test_close_plan_objective_update_failure_does_not_break_close() -> None:
    """Test that a failing objective update does not prevent plan close from succeeding."""
    # Arrange - body must include plan-header with objective_issue so
    # PlannedPRBackend._convert_to_plan() extracts objective_id correctly
    body_with_header = format_plan_header_body_for_test(objective_issue=99) + "\nPlan content"
    plan_issue = Plan(
        plan_identifier="42",
        title="Test Issue",
        body=body_with_header,
        state=PlanState.OPEN,
        url="https://github.com/owner/repo/issues/42",
        labels=["erk-pr", "erk-plan"],
        assignees=[],
        created_at=datetime(2024, 1, 1, tzinfo=UTC),
        updated_at=datetime(2024, 1, 2, tzinfo=UTC),
        metadata={"issue_body": body_with_header},
        objective_id=99,
    )

    runner = CliRunner()
    with erk_inmem_env(runner) as env:
        executor = FakePromptExecutor(command_should_fail=True)
        store, fake_github = create_plan_store_with_plans({"42": plan_issue})
        ctx = build_workspace_test_context(
            env, plan_store=store, issues=fake_github.issues, prompt_executor=executor
        )

        # Act
        result = runner.invoke(cli, ["pr", "close", "42"], obj=ctx)

        # Assert - close still succeeds even though objective update failed
        assert result.exit_code == 0
        assert "Closed plan #42" in result.output
        assert "Objective update failed" in result.output



"""Tests for ObjectiveHeaderSchema validation logic."""

import pytest

from erk_shared.gateway.github.metadata.schemas import ObjectiveHeaderSchema


class TestObjectiveHeaderSchemaValidation:
    """Test ObjectiveHeaderSchema validates required/optional fields and edge cases."""

    def test_valid_data_with_all_fields(self) -> None:
        """Valid data with all fields passes validation."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "testuser",
            "objective_comment_id": 12345,
        }
        schema.validate(data)  # Should not raise

    def test_valid_data_with_null_comment_id(self) -> None:
        """Valid data with null objective_comment_id passes."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "testuser",
            "objective_comment_id": None,
        }
        schema.validate(data)  # Should not raise

    def test_valid_data_without_optional_field(self) -> None:
        """Valid data without optional objective_comment_id passes."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "testuser",
        }
        schema.validate(data)  # Should not raise

    def test_missing_created_at(self) -> None:
        """Missing created_at raises ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_by": "testuser",
        }
        with pytest.raises(ValueError, match="Missing required fields.*created_at"):
            schema.validate(data)

    def test_missing_created_by(self) -> None:
        """Missing created_by raises ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
        }
        with pytest.raises(ValueError, match="Missing required fields.*created_by"):
            schema.validate(data)

    def test_missing_both_required_fields(self) -> None:
        """Missing both required fields raises ValueError."""
        schema = ObjectiveHeaderSchema()
        with pytest.raises(ValueError, match="Missing required fields"):
            schema.validate({})

    def test_empty_created_at(self) -> None:
        """Empty string created_at raises ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "",
            "created_by": "testuser",
        }
        with pytest.raises(ValueError, match="created_at must not be empty"):
            schema.validate(data)

    def test_empty_created_by(self) -> None:
        """Empty string created_by raises ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "",
        }
        with pytest.raises(ValueError, match="created_by must not be empty"):
            schema.validate(data)

    def test_non_string_created_at(self) -> None:
        """Non-string created_at raises ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": 12345,
            "created_by": "testuser",
        }
        with pytest.raises(ValueError, match="created_at must be a string"):
            schema.validate(data)

    def test_non_string_created_by(self) -> None:
        """Non-string created_by raises ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": 42,
        }
        with pytest.raises(ValueError, match="created_by must be a string"):
            schema.validate(data)

    def test_non_integer_comment_id(self) -> None:
        """Non-integer objective_comment_id raises ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "testuser",
            "objective_comment_id": "not-an-int",
        }
        with pytest.raises(ValueError, match="objective_comment_id must be an integer"):
            schema.validate(data)

    def test_zero_comment_id(self) -> None:
        """Zero objective_comment_id raises ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "testuser",
            "objective_comment_id": 0,
        }
        with pytest.raises(ValueError, match="objective_comment_id must be positive"):
            schema.validate(data)

    def test_negative_comment_id(self) -> None:
        """Negative objective_comment_id raises ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "testuser",
            "objective_comment_id": -1,
        }
        with pytest.raises(ValueError, match="objective_comment_id must be positive"):
            schema.validate(data)

    def test_valid_data_with_slug(self) -> None:
        """Valid data with slug field passes validation."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "testuser",
            "slug": "build-auth-system",
        }
        schema.validate(data)  # Should not raise

    def test_valid_data_with_null_slug(self) -> None:
        """Valid data with null slug passes validation."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "testuser",
            "slug": None,
        }
        schema.validate(data)  # Should not raise

    def test_empty_slug_rejected(self) -> None:
        """Empty string slug raises ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "testuser",
            "slug": "",
        }
        with pytest.raises(ValueError, match="slug must not be empty"):
            schema.validate(data)

    def test_non_string_slug_rejected(self) -> None:
        """Non-string slug raises ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "testuser",
            "slug": 42,
        }
        with pytest.raises(ValueError, match="slug must be a string"):
            schema.validate(data)

    def test_unknown_fields_rejected(self) -> None:
        """Unknown fields raise ValueError."""
        schema = ObjectiveHeaderSchema()
        data = {
            "created_at": "2025-11-25T14:37:43+00:00",
            "created_by": "testuser",
            "unexpected_field": "value",
        }
        with pytest.raises(ValueError, match="Unknown fields.*unexpected_field"):
            schema.validate(data)

    def test_get_key(self) -> None:
        """get_key returns 'objective-header'."""
        schema = ObjectiveHeaderSchema()
        assert schema.get_key() == "objective-header"



"""Tests for PlanDataProvider routing logic.

Tests verify that fetch_plans() routes queries to the correct service:
- Queries with "erk-objective" label use objective_list_service
- All other queries use plan_list_service
"""

from unittest.mock import MagicMock

from erk.tui.data.types import PlanFilters
from erk_shared.gateway.plan_data_provider.real import RealPlanDataProvider


def _make_provider(
    objective_service: MagicMock | None = None,
    plan_service: MagicMock | None = None,
) -> RealPlanDataProvider:
    """Create a RealPlanDataProvider with mocked services."""
    mock_ctx = MagicMock()
    mock_objective_service = objective_service or MagicMock()
    mock_plan_service = plan_service or MagicMock()
    mock_ctx.objective_list_service = mock_objective_service
    mock_ctx.plan_list_service = mock_plan_service
    mock_objective_service.get_objective_list_data.return_value = MagicMock(plans=[])
    mock_plan_service.get_plan_list_data.return_value = MagicMock(plans=[])

    return RealPlanDataProvider(
        mock_ctx,
        location=MagicMock(),
        clipboard=MagicMock(),
        browser=MagicMock(),
        http_client=MagicMock(),
    )


class TestPlanDataProviderRouting:
    """Tests for routing logic in RealPlanDataProvider.fetch_plans()."""

    def test_routes_to_objective_service_when_erk_objective_label_present(
        self,
    ) -> None:
        """Routes to objective_list_service when 'erk-objective' is in labels."""
        # Arrange
        mock_objective_service = MagicMock()
        mock_plan_service = MagicMock()
        mock_objective_service.get_objective_list_data.return_value = MagicMock(plans=[])
        provider = _make_provider(
            objective_service=mock_objective_service, plan_service=mock_plan_service
        )

        # Query filters with "erk-objective" label
        filters = PlanFilters(
            labels=("erk-objective",),
            state="open",
            run_state=None,
            limit=None,
            show_prs=False,
            show_runs=False,
            exclude_labels=(),
        )

        # Act
        provider.fetch_plans(filters)

        # Assert: objective_list_service should be called
        mock_objective_service.get_objective_list_data.assert_called_once()
        # plan_list_service should NOT be called
        mock_plan_service.get_plan_list_data.assert_not_called()

    def test_routes_to_plan_service_when_erk_objective_label_absent(self) -> None:
        """Routes to plan_list_service when 'erk-objective' is NOT in labels."""
        # Arrange
        mock_objective_service = MagicMock()
        mock_plan_service = MagicMock()
        provider = _make_provider(
            objective_service=mock_objective_service, plan_service=mock_plan_service
        )

        # Query filters with "erk-plan" label (NOT objective)
        filters = PlanFilters(
            labels=("erk-plan",),
            state="open",
            run_state=None,
            limit=None,
            show_prs=False,
            show_runs=False,
            exclude_labels=(),
        )

        # Act
        provider.fetch_plans(filters)

        # Assert: plan_list_service should be called
        mock_plan_service.get_plan_list_data.assert_called_once()
        # objective_list_service should NOT be called
        mock_objective_service.get_objective_list_data.assert_not_called()

    def test_routes_to_plan_service_for_learn_plans(self) -> None:
        """Routes to plan_list_service for learn plans (erk-learn label)."""
        # Arrange
        mock_objective_service = MagicMock()
        mock_plan_service = MagicMock()
        provider = _make_provider(
            objective_service=mock_objective_service, plan_service=mock_plan_service
        )

        # Query filters with "erk-learn" label
        filters = PlanFilters(
            labels=("erk-learn",),
            state="open",
            run_state=None,
            limit=None,
            show_prs=False,
            show_runs=False,
            exclude_labels=(),
        )

        # Act
        provider.fetch_plans(filters)

        # Assert: plan_list_service should be called (not objective service)
        mock_plan_service.get_plan_list_data.assert_called_once()
        mock_objective_service.get_objective_list_data.assert_not_called()

    def test_routes_to_plan_service_for_multi_label_query(self) -> None:
        """Routes to plan_list_service for multi-label queries (erk-pr + erk-learn)."""
        # Arrange
        mock_objective_service = MagicMock()
        mock_plan_service = MagicMock()
        provider = _make_provider(
            objective_service=mock_objective_service, plan_service=mock_plan_service
        )

        # Query filters with multiple labels, but NOT erk-objective
        filters = PlanFilters(
            labels=("erk-pr", "erk-learn"),
            state="open",
            run_state=None,
            limit=None,
            show_prs=False,
            show_runs=False,
            exclude_labels=(),
        )

        # Act
        provider.fetch_plans(filters)

        # Assert: plan_list_service should be called
        mock_plan_service.get_plan_list_data.assert_called_once()
        mock_objective_service.get_objective_list_data.assert_not_called()

    def test_condition_inverts_correctly_objective_vs_plan(self) -> None:
        """Verify the routing condition inverts correctly for objectives vs plans.

        This test explicitly checks the behavior documented in the routing logic:
        - if "erk-objective" in labels → use objective service (inverted check)
        - else → use plan service (all other queries)
        """
        # Arrange
        mock_objective_service = MagicMock()
        mock_plan_service = MagicMock()
        provider = _make_provider(
            objective_service=mock_objective_service, plan_service=mock_plan_service
        )

        # Case 1: WITH erk-objective
        filters_with_objective = PlanFilters(
            labels=("erk-objective",),
            state="open",
            run_state=None,
            limit=None,
            show_prs=False,
            show_runs=False,
            exclude_labels=(),
        )
        provider.fetch_plans(filters_with_objective)
        assert mock_objective_service.get_objective_list_data.call_count == 1, (
            "Should route WITH objective to objective service"
        )
        assert mock_plan_service.get_plan_list_data.call_count == 0, (
            "Should NOT route WITH objective to plan service"
        )

        # Reset mocks
        mock_objective_service.reset_mock()
        mock_plan_service.reset_mock()

        # Case 2: WITHOUT erk-objective (plan query)
        filters_without_objective = PlanFilters(
            labels=("erk-plan",),
            state="open",
            run_state=None,
            limit=None,
            show_prs=False,
            show_runs=False,
            exclude_labels=(),
        )
        provider.fetch_plans(filters_without_objective)
        assert mock_plan_service.get_plan_list_data.call_count == 1, (
            "Should route WITHOUT objective to plan service"
        )
        assert mock_objective_service.get_objective_list_data.call_count == 0, (
            "Should NOT route WITHOUT objective to objective service"
        )



"""Unit tests for plan-update command."""

import json
from datetime import UTC, datetime

from click.testing import CliRunner

from erk.cli.commands.exec.scripts.plan_update import plan_update
from erk_shared.context.context import ErkContext
from erk_shared.gateway.claude_installation.fake import FakeClaudeInstallation
from erk_shared.gateway.git.fake import FakeGit
from erk_shared.gateway.git.remote_ops.types import PushError
from erk_shared.gateway.github.fake import FakeGitHub
from erk_shared.gateway.github.issues.fake import FakeGitHubIssues
from erk_shared.gateway.github.issues.types import IssueComment, IssueInfo
from erk_shared.gateway.github.types import PRDetails
from erk_shared.gateway.time.fake import FakeTime
from erk_shared.plan_store.planned_pr import PlannedPRBackend
from erk_shared.plan_store.planned_pr_lifecycle import IMPL_CONTEXT_DIR
from tests.test_utils.plan_helpers import (
    format_plan_header_body_for_test,
    issue_info_to_pr_details,
)


def _make_issue(
    number: int,
    title: str,
    body: str,
) -> IssueInfo:
    """Create a test IssueInfo."""
    now = datetime.now(UTC)
    return IssueInfo(
        number=number,
        title=title,
        body=body,
        state="OPEN",
        url=f"https://github.com/test/repo/issues/{number}",
        labels=["erk-pr", "erk-plan"],
        assignees=[],
        created_at=now,
        updated_at=now,
        author="testuser",
    )


def _make_comment(comment_id: int, body: str) -> IssueComment:
    """Create a test IssueComment."""
    return IssueComment(
        body=body,
        url=f"https://github.com/test/repo/issues/1#issuecomment-{comment_id}",
        id=comment_id,
        author="testuser",
    )


def test_plan_update_success() -> None:
    """Test successful plan update."""
    issue = _make_issue(42, "Test Plan [erk-plan]", "metadata body")
    comment = _make_comment(12345, "old plan content")
    fake_gh = FakeGitHubIssues(
        issues={42: issue},
        comments_with_urls={42: [comment]},
    )
    fake_github = FakeGitHub(
        pr_details={42: issue_info_to_pr_details(issue)},
        issues_gateway=fake_gh,
    )
    plan_content = """# Updated Plan

- New step 1
- New step 2"""
    fake_store = FakeClaudeInstallation.for_test(plans={"test-plan": plan_content})
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "42", "--format", "json"],
        obj=ErkContext.for_test(
            github=fake_github,
            plan_store=PlannedPRBackend(fake_github, fake_gh, time=FakeTime()),
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 0, f"Failed: {result.output}"
    output = json.loads(result.output)
    assert output["success"] is True
    assert output["plan_number"] == 42

    assert output["title"] == "[erk-plan] Updated Plan"

    # Verify update_plan_content was called (updates PR body with PlannedPRBackend)
    content_bodies = [b for n, b in fake_github.updated_pr_bodies if n == 42]
    assert len(content_bodies) >= 1
    updated_body = content_bodies[-1]
    assert "New step 1" in updated_body
    assert "New step 2" in updated_body

    # Verify update_plan_title was called
    assert len(fake_github.updated_pr_titles) == 1
    assert fake_github.updated_pr_titles[0] == (42, "[erk-plan] Updated Plan")


def test_plan_update_display_format() -> None:
    """Test display output format."""
    issue = _make_issue(99, "My Feature [erk-plan]", "body")
    comment = _make_comment(55555, "plan content")
    fake_gh = FakeGitHubIssues(
        issues={99: issue},
        comments_with_urls={99: [comment]},
    )
    fake_github = FakeGitHub(
        pr_details={99: issue_info_to_pr_details(issue)},
        issues_gateway=fake_gh,
    )
    plan_content = """# Display Test

- Step"""
    fake_store = FakeClaudeInstallation.for_test(plans={"display-test": plan_content})
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "99", "--format", "display"],
        obj=ErkContext.for_test(
            github=fake_github,
            plan_store=PlannedPRBackend(fake_github, fake_gh, time=FakeTime()),
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 0
    assert "Plan #99 updated" in result.output
    assert "Title: [erk-plan] Display Test" in result.output
    assert "URL: " in result.output


def test_plan_update_no_plan_found() -> None:
    """Test error when no plan found."""
    issue = _make_issue(42, "Test [erk-plan]", "body")
    comment = _make_comment(12345, "plan content")
    fake_gh = FakeGitHubIssues(
        issues={42: issue},
        comments_with_urls={42: [comment]},
    )
    fake_github = FakeGitHub(
        pr_details={42: issue_info_to_pr_details(issue)},
        issues_gateway=fake_gh,
    )
    # Empty store - no plans
    fake_store = FakeClaudeInstallation.for_test()
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "42", "--format", "json"],
        obj=ErkContext.for_test(
            github=fake_github,
            plan_store=PlannedPRBackend(fake_github, fake_gh, time=FakeTime()),
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 1
    output = json.loads(result.output)
    assert output["success"] is False
    assert "No plan found" in output["error"]


def test_plan_update_issue_not_found() -> None:
    """Test error when issue does not exist."""
    # Empty issues dict - no issues
    fake_gh = FakeGitHubIssues()
    fake_github = FakeGitHub(issues_gateway=fake_gh)
    plan_content = """# Test

- Step"""
    fake_store = FakeClaudeInstallation.for_test(plans={"test": plan_content})
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "999", "--format", "json"],
        obj=ErkContext.for_test(
            github=fake_github,
            plan_store=PlannedPRBackend(fake_github, fake_gh, time=FakeTime()),
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 1
    output = json.loads(result.output)
    assert output["success"] is False
    assert "999" in output["error"]


def test_plan_update_formats_plan_content() -> None:
    """Test that plan content is properly formatted with metadata block."""
    issue = _make_issue(42, "Test [erk-plan]", "body")
    comment = _make_comment(12345, "old content")
    fake_gh = FakeGitHubIssues(
        issues={42: issue},
        comments_with_urls={42: [comment]},
    )
    fake_github = FakeGitHub(
        pr_details={42: issue_info_to_pr_details(issue)},
        issues_gateway=fake_gh,
    )
    plan_content = """# My Plan

- Step 1"""
    fake_store = FakeClaudeInstallation.for_test(plans={"format-test": plan_content})
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "42"],
        obj=ErkContext.for_test(
            github=fake_github,
            plan_store=PlannedPRBackend(fake_github, fake_gh, time=FakeTime()),
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 0

    # Verify the plan content was stored in the PR body
    content_bodies = [b for n, b in fake_github.updated_pr_bodies if n == 42]
    assert len(content_bodies) >= 1
    updated_body = content_bodies[-1]
    # PlannedPRBackend stores content directly (no comment wrapper)
    assert "# My Plan" in updated_body
    assert "Step 1" in updated_body


def test_plan_update_updates_title_from_plan() -> None:
    """Test that issue title is updated from plan H1 heading."""
    issue = _make_issue(42, "[erk-plan] Old Title", "body")
    comment = _make_comment(12345, "old content")
    fake_gh = FakeGitHubIssues(
        issues={42: issue},
        comments_with_urls={42: [comment]},
    )
    fake_github = FakeGitHub(
        pr_details={42: issue_info_to_pr_details(issue)},
        issues_gateway=fake_gh,
    )
    plan_content = """# New Feature Name

- Step 1
- Step 2"""
    fake_store = FakeClaudeInstallation.for_test(plans={"title-test": plan_content})
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "42", "--format", "json"],
        obj=ErkContext.for_test(
            github=fake_github,
            plan_store=PlannedPRBackend(fake_github, fake_gh, time=FakeTime()),
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 0, f"Failed: {result.output}"
    output = json.loads(result.output)
    assert output["title"] == "[erk-plan] New Feature Name"

    # Verify update_plan_title was called
    assert len(fake_github.updated_pr_titles) == 1
    assert fake_github.updated_pr_titles[0] == (42, "[erk-plan] New Feature Name")


def test_plan_update_learn_plan_gets_learn_tag() -> None:
    """Test that learn plans get [erk-learn] title tag."""
    now = datetime.now(UTC)
    issue = IssueInfo(
        number=42,
        title="[erk-learn] Old Title",
        body="body",
        state="OPEN",
        url="https://github.com/test/repo/issues/42",
        labels=["erk-pr", "erk-learn"],
        assignees=[],
        created_at=now,
        updated_at=now,
        author="testuser",
    )
    comment = _make_comment(12345, "old content")
    fake_gh = FakeGitHubIssues(
        issues={42: issue},
        comments_with_urls={42: [comment]},
    )
    fake_github = FakeGitHub(
        pr_details={42: issue_info_to_pr_details(issue)},
        issues_gateway=fake_gh,
    )
    plan_content = """# Learn Something New

- Insight 1"""
    fake_store = FakeClaudeInstallation.for_test(plans={"learn-test": plan_content})
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "42", "--format", "json"],
        obj=ErkContext.for_test(
            github=fake_github,
            plan_store=PlannedPRBackend(fake_github, fake_gh, time=FakeTime()),
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 0, f"Failed: {result.output}"
    output = json.loads(result.output)
    assert output["title"] == "[erk-learn] Learn Something New"

    assert fake_github.updated_pr_titles[0] == (42, "[erk-learn] Learn Something New")


def test_plan_update_strips_plan_prefix_from_title() -> None:
    """Test that 'Plan: ' prefix is stripped from extracted title."""
    issue = _make_issue(42, "[erk-plan] Old Title", "body")
    comment = _make_comment(12345, "old content")
    fake_gh = FakeGitHubIssues(
        issues={42: issue},
        comments_with_urls={42: [comment]},
    )
    fake_github = FakeGitHub(
        pr_details={42: issue_info_to_pr_details(issue)},
        issues_gateway=fake_gh,
    )
    plan_content = """# Plan: Add Feature X

- Step 1"""
    fake_store = FakeClaudeInstallation.for_test(plans={"prefix-test": plan_content})
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "42", "--format", "json"],
        obj=ErkContext.for_test(
            github=fake_github,
            plan_store=PlannedPRBackend(fake_github, fake_gh, time=FakeTime()),
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 0, f"Failed: {result.output}"
    output = json.loads(result.output)
    assert output["title"] == "[erk-plan] Add Feature X"


def test_plan_update_display_format_shows_new_title() -> None:
    """Test display format shows the updated title."""
    issue = _make_issue(42, "[erk-plan] Old Title", "body")
    comment = _make_comment(12345, "old content")
    fake_gh = FakeGitHubIssues(
        issues={42: issue},
        comments_with_urls={42: [comment]},
    )
    fake_github = FakeGitHub(
        pr_details={42: issue_info_to_pr_details(issue)},
        issues_gateway=fake_gh,
    )
    plan_content = """# Updated Feature

- Step 1"""
    fake_store = FakeClaudeInstallation.for_test(plans={"display-title": plan_content})
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "42", "--format", "display"],
        obj=ErkContext.for_test(
            github=fake_github,
            plan_store=PlannedPRBackend(fake_github, fake_gh, time=FakeTime()),
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 0
    assert "Title: [erk-plan] Updated Feature" in result.output


# ============================================================================
# Branch push tests (PR-backed plans with PlannedPRBackend)
# ============================================================================


def _make_pr_details(
    number: int,
    title: str,
    *,
    branch_name: str,
) -> PRDetails:
    """Create PRDetails with a plan-header metadata block containing branch_name."""
    from erk_shared.plan_store.planned_pr_lifecycle import build_plan_stage_body

    metadata_body = format_plan_header_body_for_test(branch_name=branch_name)
    plan_content = "# Old Plan Content"
    pr_body = build_plan_stage_body(metadata_body, plan_content, summary=None)

    return PRDetails(
        number=number,
        url=f"https://github.com/test/repo/pull/{number}",
        title=title,
        body=pr_body,
        state="OPEN",
        is_draft=True,
        base_ref_name="master",
        head_ref_name=branch_name,
        is_cross_repository=False,
        mergeable="UNKNOWN",
        merge_state_status="UNKNOWN",
        owner="test-owner",
        repo="test-repo",
        labels=("erk-plan",),
    )


def test_plan_update_pushes_to_branch() -> None:
    """Test that plan update pushes to branch when branch_name is in header_fields."""
    branch_name = "plnd/my-feature-branch"
    pr = _make_pr_details(42, "[erk-plan] Old Title", branch_name=branch_name)
    fake_github = FakeGitHub(pr_details={42: pr})
    fake_git = FakeGit()
    plan_content = """# Updated Plan

- New step 1
- New step 2"""
    fake_store = FakeClaudeInstallation.for_test(plans={"test-plan": plan_content})
    backend = PlannedPRBackend(fake_github, fake_github.issues, time=FakeTime())
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "42", "--format", "json"],
        obj=ErkContext.for_test(
            github=fake_github,
            git=fake_git,
            plan_store=backend,
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 0, f"Failed: {result.output}"
    output = json.loads(result.output)
    assert output["success"] is True
    assert output["branch_name"] == branch_name
    assert output["branch_updated"] is True

    # Verify commit_files_to_branch was called with correct args
    assert len(fake_git.branch_commits) == 1
    commit = fake_git.branch_commits[0]
    assert commit.branch == branch_name
    assert f"{IMPL_CONTEXT_DIR}/plan.md" in commit.files
    assert "New step 1" in commit.files[f"{IMPL_CONTEXT_DIR}/plan.md"]

    # Verify push_to_remote was called
    assert len(fake_git.pushed_branches) == 1
    push = fake_git.pushed_branches[0]
    assert push.remote == "origin"
    assert push.branch == branch_name
    assert push.set_upstream is False
    assert push.force is False


def test_plan_update_branch_push_failure_still_succeeds() -> None:
    """Test that push failure still reports success (PR body was updated)."""
    branch_name = "plnd/my-feature-branch"
    pr = _make_pr_details(42, "[erk-plan] Old Title", branch_name=branch_name)
    fake_github = FakeGitHub(pr_details={42: pr})
    fake_git = FakeGit(push_to_remote_error=PushError(message="rejected"))
    plan_content = """# Updated Plan

- Step 1"""
    fake_store = FakeClaudeInstallation.for_test(plans={"test-plan": plan_content})
    backend = PlannedPRBackend(fake_github, fake_github.issues, time=FakeTime())
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "42", "--format", "json"],
        obj=ErkContext.for_test(
            github=fake_github,
            git=fake_git,
            plan_store=backend,
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 0, f"Failed: {result.output}"
    output = json.loads(result.output)
    assert output["success"] is True
    assert output["branch_name"] == branch_name
    assert output["branch_updated"] is False

    # Commit was still made (before push failed)
    assert len(fake_git.branch_commits) == 1


def test_plan_update_no_branch_skips_push() -> None:
    """Test that issue-backed plans (no branch_name) skip branch push entirely."""
    issue = _make_issue(42, "[erk-plan] Old Title", "body")
    comment = _make_comment(12345, "old content")
    fake_gh = FakeGitHubIssues(
        issues={42: issue},
        comments_with_urls={42: [comment]},
    )
    fake_github = FakeGitHub(
        pr_details={42: issue_info_to_pr_details(issue)},
        issues_gateway=fake_gh,
    )
    fake_git = FakeGit()
    plan_content = """# Updated Plan

- Step 1"""
    fake_store = FakeClaudeInstallation.for_test(plans={"test-plan": plan_content})
    runner = CliRunner()

    result = runner.invoke(
        plan_update,
        ["--plan-number", "42", "--format", "json"],
        obj=ErkContext.for_test(
            github=fake_github,
            git=fake_git,
            plan_store=PlannedPRBackend(fake_github, fake_gh, time=FakeTime()),
            claude_installation=fake_store,
        ),
    )

    assert result.exit_code == 0, f"Failed: {result.output}"
    output = json.loads(result.output)
    assert output["success"] is True
    assert output["branch_name"] is None
    assert output["branch_updated"] is False

    # No git operations attempted
    assert len(fake_git.branch_commits) == 0
    assert len(fake_git.pushed_branches) == 0



from datetime import datetime

from erk_shared.gateway.time.fake import DEFAULT_FAKE_TIME, FakeTime


def test_now_returns_default_time() -> None:
    fake_time = FakeTime()
    assert fake_time.now() == DEFAULT_FAKE_TIME


def test_now_returns_custom_time() -> None:
    custom = datetime(2025, 6, 15, 10, 0, 0)
    fake_time = FakeTime(current_time=custom)
    assert fake_time.now() == custom


def test_sleep_tracks_calls() -> None:
    fake_time = FakeTime()
    fake_time.sleep(1.5)
    fake_time.sleep(0.5)
    assert fake_time.sleep_calls == [1.5, 0.5]


def test_sleep_calls_initially_empty() -> None:
    fake_time = FakeTime()
    assert fake_time.sleep_calls == []


def test_monotonic_returns_default_zero() -> None:
    fake_time = FakeTime()
    assert fake_time.monotonic() == 0.0


def test_monotonic_returns_fixed_value() -> None:
    fake_time = FakeTime(monotonic_values=[100.0])
    assert fake_time.monotonic() == 100.0
    assert fake_time.monotonic() == 100.0


def test_monotonic_returns_sequence_in_order() -> None:
    fake_time = FakeTime(monotonic_values=[1.0, 2.0, 3.0])
    assert fake_time.monotonic() == 1.0
    assert fake_time.monotonic() == 2.0
    assert fake_time.monotonic() == 3.0


def test_monotonic_repeats_last_value_when_exhausted() -> None:
    fake_time = FakeTime(monotonic_values=[1.0, 5.0])
    assert fake_time.monotonic() == 1.0
    assert fake_time.monotonic() == 5.0
    assert fake_time.monotonic() == 5.0
    assert fake_time.monotonic() == 5.0



"""Tests for next_steps formatting."""

from erk_shared.output.next_steps import (
    IssueNextSteps,
    PlannedPRNextSteps,
    format_next_steps_plain,
    format_planned_pr_next_steps_plain,
)


def test_planned_pr_next_steps_constructs_with_all_params() -> None:
    """PlannedPRNextSteps constructs correctly with pr_number, branch_name, and url."""
    s = PlannedPRNextSteps(
        pr_number=42, branch_name="plan-feature-foo", url="https://github.com/org/repo/pull/42"
    )
    assert s.pr_number == 42
    assert s.branch_name == "plan-feature-foo"
    assert s.url == "https://github.com/org/repo/pull/42"


def test_planned_pr_next_steps_view_returns_url() -> None:
    """view returns the URL directly."""
    s = PlannedPRNextSteps(
        pr_number=42, branch_name="plan-feature-foo", url="https://github.com/org/repo/pull/42"
    )
    assert s.view == "https://github.com/org/repo/pull/42"


def test_planned_pr_next_steps_checkout_branch_and_implement_uses_branch_name() -> None:
    """checkout_branch_and_implement uses branch_name, not pr_number."""
    s = PlannedPRNextSteps(
        pr_number=42, branch_name="plan-feature-foo", url="https://github.com/org/repo/pull/42"
    )
    assert "plan-feature-foo" in s.checkout_branch_and_implement
    assert "42" not in s.checkout_branch_and_implement


def test_planned_pr_next_steps_checkout_uses_pr_number() -> None:
    """checkout uses --for-plan with pr_number."""
    s = PlannedPRNextSteps(
        pr_number=42, branch_name="plan-feature-foo", url="https://github.com/org/repo/pull/42"
    )
    assert s.checkout == "erk br co --for-plan 42"


def test_issue_next_steps_checkout_uses_co() -> None:
    """IssueNextSteps.checkout uses erk br co --for-plan."""
    s = IssueNextSteps(plan_number=99, url="https://github.com/org/repo/issues/99")
    assert s.checkout == "erk br co --for-plan 99"


def test_issue_next_steps_view_returns_url() -> None:
    """IssueNextSteps.view returns the URL directly."""
    s = IssueNextSteps(plan_number=99, url="https://github.com/org/repo/issues/99")
    assert s.view == "https://github.com/org/repo/issues/99"


def test_format_planned_pr_next_steps_plain_uses_for_plan() -> None:
    """format_planned_pr_next_steps_plain uses --for-plan command."""
    output = format_planned_pr_next_steps_plain(
        42, branch_name="plan-feature-foo", url="https://github.com/org/repo/pull/42"
    )
    assert "erk br co --for-plan 42" in output


def test_format_planned_pr_next_steps_plain_contains_url() -> None:
    """format_planned_pr_next_steps_plain shows URL in view line."""
    output = format_planned_pr_next_steps_plain(
        42, branch_name="plan-feature-foo", url="https://github.com/org/repo/pull/42"
    )
    assert "https://github.com/org/repo/pull/42" in output


def test_format_next_steps_plain_uses_co() -> None:
    """format_next_steps_plain uses erk br co --for-plan."""
    output = format_next_steps_plain(99, url="https://github.com/org/repo/issues/99")
    assert "erk br co --for-plan 99" in output
    assert "erk br create" not in output


def test_format_next_steps_plain_contains_implement() -> None:
    """format_next_steps_plain includes implement command."""
    output = format_next_steps_plain(99, url="https://github.com/org/repo/issues/99")
    assert "erk implement --dangerous" in output


def test_format_next_steps_plain_contains_url() -> None:
    """format_next_steps_plain shows URL in view line."""
    output = format_next_steps_plain(99, url="https://github.com/org/repo/issues/99")
    assert "https://github.com/org/repo/issues/99" in output


def test_format_planned_pr_next_steps_plain_contains_implement() -> None:
    """format_planned_pr_next_steps_plain includes implement command."""
    output = format_planned_pr_next_steps_plain(
        42, branch_name="plan-feature-foo", url="https://github.com/org/repo/pull/42"
    )
    assert "erk implement --dangerous" in output



"""Pure parsing functions for GitHub PR and workflow data.

Extracted from RealGitHub to enable reuse by both the subprocess path
(via RealGitHub methods) and the direct HTTP path (via PlannedPRPlanListService).

All functions are stateless — they transform data without side effects.
"""

from datetime import datetime
from typing import Any

from erk_shared.gateway.github.parsing import parse_aggregated_check_counts
from erk_shared.gateway.github.types import (
    BRANCH_NOT_AVAILABLE,
    DISPLAY_TITLE_NOT_AVAILABLE,
    GitHubRepoId,
    MergeableStatus,
    PRDetails,
    PullRequestInfo,
    StatusCheckRollupData,
    WorkflowRun,
    WorkflowRunConclusion,
    WorkflowRunStatus,
)


def parse_status_rollup(
    status_rollup: StatusCheckRollupData | None,
) -> tuple[bool | None, tuple[int, int] | None]:
    """Parse checks status and counts from statusCheckRollup.

    Returns (checks_passing, checks_counts).
    """
    if status_rollup is None:
        return (None, None)

    rollup_state = status_rollup.get("state")
    checks_passing = None
    if rollup_state == "SUCCESS":
        checks_passing = True
    elif rollup_state in ("FAILURE", "ERROR"):
        checks_passing = False

    checks_counts = None
    contexts = status_rollup.get("contexts")
    if contexts is not None and isinstance(contexts, dict):
        total = contexts.get("totalCount", 0)
        if total > 0:
            checks_counts = parse_aggregated_check_counts(
                contexts.get("checkRunCountsByState", []),
                contexts.get("statusContextCountsByState", []),
                total,
            )

    return (checks_passing, checks_counts)


def parse_mergeable_status(mergeable: MergeableStatus | None) -> bool | None:
    """Parse has_conflicts from mergeable field."""
    if mergeable == "CONFLICTING":
        return True
    if mergeable == "MERGEABLE":
        return False
    return None


def parse_review_thread_counts(
    review_threads: dict[str, Any] | None,
) -> tuple[int, int] | None:
    """Parse review thread counts from reviewThreads field.

    Returns (resolved_count, total_count) or None if not available.
    """
    if review_threads is None:
        return None

    total_count = review_threads.get("totalCount", 0)
    if total_count == 0:
        return (0, 0)

    nodes = review_threads.get("nodes", [])
    resolved_count = sum(1 for node in nodes if node and node.get("isResolved", False))

    return (resolved_count, total_count)


def merge_rest_graphql_pr_data(
    rest_items: list[dict[str, Any]],
    enrichment: dict[int, dict[str, Any]],
    repo_id: GitHubRepoId,
) -> tuple[list[PRDetails], dict[int, list[PullRequestInfo]], int]:
    """Merge REST issue data with GraphQL PR enrichment into PRDetails and PullRequestInfo.

    Args:
        rest_items: Raw REST API issue/PR items
        enrichment: GraphQL enrichment data keyed by PR number
        repo_id: GitHub repository identity

    Returns:
        Tuple of (pr_details_list, pr_linkages_by_pr_number, unenriched_count).
        PRs without GraphQL enrichment data are included in pr_details_list
        (needed for plan conversion) but excluded from pr_linkages to avoid
        misleading fallback defaults for branch, draft status, and checks.
    """
    pr_details_list: list[PRDetails] = []
    pr_linkages: dict[int, list[PullRequestInfo]] = {}
    unenriched_count = 0

    for item in rest_items:
        pr_number = item["number"]
        gql = enrichment.get(pr_number, {})

        rest_head_ref = item.get("pull_request", {}).get("head", {}).get("ref", "")
        head_ref_name = gql.get("headRefName", rest_head_ref)
        base_ref_name = gql.get("baseRefName", "")
        pr_state = item.get("state", "open").upper()

        pr_details = PRDetails(
            number=pr_number,
            url=item.get("html_url", ""),
            title=item.get("title", ""),
            body=item.get("body", "") or "",
            state=pr_state,
            is_draft=gql.get("isDraft", False),
            base_ref_name=base_ref_name,
            head_ref_name=head_ref_name,
            is_cross_repository=gql.get("isCrossRepository", False),
            mergeable=gql.get("mergeable", "UNKNOWN"),
            merge_state_status=gql.get("mergeStateStatus", "UNKNOWN"),
            owner=repo_id.owner,
            repo=repo_id.repo,
            labels=tuple(label["name"] for label in item.get("labels", [])),
            created_at=datetime.fromisoformat(item["created_at"].replace("Z", "+00:00")),
            updated_at=datetime.fromisoformat(item["updated_at"].replace("Z", "+00:00")),
            author=item.get("user", {}).get("login", ""),
        )
        pr_details_list.append(pr_details)

        # Only create PullRequestInfo when enrichment data is available.
        # Without GraphQL data, fields like base_ref_name, is_draft, and checks
        # would have misleading fallback defaults.
        if not gql:
            unenriched_count += 1
            continue

        # Build PullRequestInfo with rich GraphQL data
        checks_passing, checks_counts = parse_status_rollup(gql.get("statusCheckRollup"))
        has_conflicts = parse_mergeable_status(gql.get("mergeable"))
        review_thread_counts = parse_review_thread_counts(gql.get("reviewThreads"))
        review_decision = gql.get("reviewDecision")

        pr_info = PullRequestInfo(
            number=pr_number,
            state=pr_state,
            url=item.get("html_url", ""),
            is_draft=pr_details.is_draft,
            title=item.get("title"),
            checks_passing=checks_passing,
            owner=repo_id.owner,
            repo=repo_id.repo,
            has_conflicts=has_conflicts,
            checks_counts=checks_counts,
            review_thread_counts=review_thread_counts,
            head_branch=head_ref_name,
            review_decision=review_decision,
            base_ref_name=base_ref_name,
        )
        pr_linkages[pr_number] = [pr_info]

    return (pr_details_list, pr_linkages, unenriched_count)


def parse_workflow_runs_nodes_response(
    response: dict[str, Any],
    node_ids: list[str],
) -> dict[str, WorkflowRun | None]:
    """Parse GraphQL nodes response into WorkflowRun objects.

    Maps the GraphQL checkSuite status/conclusion to WorkflowRun fields.

    Args:
        response: GraphQL response data
        node_ids: Original list of node IDs (for result ordering)

    Returns:
        Mapping of node_id -> WorkflowRun or None
    """
    status_map: dict[str | None, WorkflowRunStatus] = {
        "COMPLETED": "completed",
        "IN_PROGRESS": "in_progress",
        "QUEUED": "queued",
    }
    conclusion_map: dict[str | None, WorkflowRunConclusion] = {
        "SUCCESS": "success",
        "FAILURE": "failure",
        "SKIPPED": "skipped",
        "CANCELLED": "cancelled",
    }

    result: dict[str, WorkflowRun | None] = {}
    nodes = response.get("data", {}).get("nodes", [])

    for node in nodes:
        if node is None:
            continue

        node_id = node.get("id")
        if node_id is None:
            continue

        check_suite = node.get("checkSuite")
        status: WorkflowRunStatus | None = None
        conclusion: WorkflowRunConclusion | None = None
        head_sha: str | None = None

        if check_suite is not None:
            status = status_map.get(check_suite.get("status"))
            conclusion = conclusion_map.get(check_suite.get("conclusion"))

            commit = check_suite.get("commit")
            if commit is not None:
                head_sha = commit.get("oid")

        # Parse created_at timestamp
        created_at = None
        created_at_str = node.get("createdAt")
        if created_at_str:
            created_at = datetime.fromisoformat(created_at_str.replace("Z", "+00:00"))

        workflow_run = WorkflowRun(
            run_id=str(node.get("databaseId", "")),
            status=status or "unknown",  # Default for missing status
            conclusion=conclusion,
            branch=BRANCH_NOT_AVAILABLE,
            head_sha=head_sha or "",  # Default for missing SHA
            display_title=DISPLAY_TITLE_NOT_AVAILABLE,
            created_at=created_at,
        )
        result[node_id] = workflow_run

    # Ensure all requested node_ids are in result (with None for missing)
    for node_id in node_ids:
        if node_id not in result:
            result[node_id] = None

    return result
